@@ -571,7 +571,7 @@ def generate(grm: util.PathLike,
 
 # The following defines the command-line options available for users to
 # specify in ACEProcess tasks. For a description of these options, see:
-#     http://moin.delph-in.net/AceOptions
+#     https://github.com/delph-in/docs/wiki/AceOptions
 
 # thanks: https://stackoverflow.com/a/14728477/1441112
 class _ACEArgumentParser(argparse.ArgumentParser):
@@ -668,7 +668,7 @@ def _make_response(lines, run) -> Tuple[interface.Response, List[str]]:
     return response, content_lines
 
 
-def _sexpr_data(line: str) -> Iterator[Dict[str, Any]]:
+def _sexpr_data(line: str) -> Iterator[Tuple[str, Any]]:
     while line:
         try:
             expr = util.SExpr.parse(line)
@@ -677,10 +677,14 @@ def _sexpr_data(line: str) -> Iterator[Dict[str, Any]]:
                 (':error', 'incomplete output from ACE'),
                 '')
         if len(expr.data) != 2:
-            logger.error('Malformed output from ACE: %s', line)
+            logger.error('Could not read output from ACE: %s', line)
             break
+
+        key, val = expr.data
+        assert isinstance(key, str)
+        yield key, val
+
         line = expr.remainder.lstrip()
-        yield expr.data
 
 
 def _tsdb_response(response: interface.Response,
