@@ -9,7 +9,11 @@ Dependency Minimal Recursion Semantics ([DMRS]_)
   Association for Computational Linguistics, 2009.
 """
 
-from delphin.dmrs._exceptions import DMRSError, DMRSSyntaxError
+from delphin.dmrs._exceptions import (
+    DMRSError,
+    DMRSSyntaxError,
+    DMRSWarning,
+)
 from delphin.dmrs._dmrs import (
     DMRS,
     Node,
@@ -21,7 +25,8 @@ from delphin.dmrs._dmrs import (
     HEQ_POST,
     NEQ_POST,
     H_POST,
-    CVARSORT)
+    CVARSORT,
+)
 from delphin.dmrs._operations import from_mrs
 # Default modules need to import the PyDelphin version
 from delphin.__about__ import __version__  # noqa: F401
@@ -42,4 +47,5 @@ __all__ = [
     'from_mrs',
     'DMRSError',
     'DMRSSyntaxError',
+    'DMRSWarning',
 ]
