@@ -21,6 +21,7 @@ from typing import (
 from sre_parse import parse_template
 from pathlib import Path
 from array import array
+from itertools import takewhile
 import warnings
 import logging
 
@@ -53,6 +54,11 @@ if TYPE_CHECKING:
 else:
     _CMap = array
 
+# Mask values
+_MASK_B = 1  # start of mask
+_MASK_I = 2  # inside mask
+_MASK_O = 0  # not masked
+
 
 class REPPError(PyDelphinException):
     """Raised when there is an error in tokenizing with REPP."""
@@ -95,6 +101,7 @@ class REPPStep(NamedTuple):
         applied (bool): `True` if the rule was applied
         startmap (:py:class:`array`): integer array of start offsets
         endmap (:py:class:`array`): integer array of end offsets
+        mask (:py:class:`array`): integer array of mask indicators
     """
     input: str
     output: str
@@ -102,75 +109,28 @@ class REPPStep(NamedTuple):
     applied: bool
     startmap: _CMap
     endmap: _CMap
+    mask: _CMap  # BIO scheme, B=1, I=2, O=0
 
 
 _Trace = Iterator[Union[REPPStep, REPPResult]]
 
 
-class _REPPOperation(object):
+class _REPPOperation:
     """
     The supertype of REPP groups and rules.
 
     This class defines the apply(), trace(), and tokenize() methods
     which are available in [_REPPRule], [_REPPGroup],
-    [_REPPIterativeGroup], and [REPP] instances.
+    [_REPPInternalGroup], and [REPP] instances.
     """
-    def _apply(self, s: str, active: Set[str]) -> Iterator[REPPStep]:
-        raise NotImplementedError()
 
-    def apply(self, s: str, active: Iterable[str] = None) -> REPPResult:
-        logger.info('apply(%r)', s)
-        for step in self._trace(s, set(active or []), False):
-            pass  # we only care about the last step
-        assert isinstance(step, REPPResult)
-        return step
-
-    def trace(
-        self, s: str, active: Iterable[str] = None, verbose: bool = False
-    ) -> _Trace:
-        logger.info('trace(%r)', s)
-        yield from self._trace(s, set(active or []), verbose)
-
-    def _trace(
-        self, s: str, active: Set[str], verbose: bool
-    ) -> _Trace:
-        startmap = _zeromap(s)
-        endmap = _zeromap(s)
-        # initial boundaries
-        startmap[0] = 1
-        endmap[-1] = -1
-        step = None
-        for step in self._apply(s, active):
-            if step.applied or verbose:
-                yield step
-            if step.applied:
-                startmap = _mergemap(startmap, step.startmap)
-                endmap = _mergemap(endmap, step.endmap)
-        if step is not None:
-            s = step.output
-        yield REPPResult(s, startmap, endmap)
-
-    def tokenize(
+    def _apply(
         self,
         s: str,
-        pattern: str = DEFAULT_TOKENIZER,
-        active: Iterable[str] = None
-    ) -> YYTokenLattice:
-        logger.info('tokenize(%r, %r)', s, pattern)
-        res = self.apply(s, active=set(active or []))
-        return self.tokenize_result(res, pattern=pattern)
-
-    def tokenize_result(
-            self, result: REPPResult, pattern: str = DEFAULT_TOKENIZER
-    ) -> YYTokenLattice:
-        logger.info('tokenize_result(%r, %r)', result, pattern)
-        tokens = [
-            YYToken(id=i, start=i, end=(i + 1),
-                    lnk=Lnk.charspan(tok[0], tok[1]),
-                    form=tok[2])
-            for i, tok in enumerate(_tokenize(result, pattern))
-        ]
-        return YYTokenLattice(tokens)
+        active: Set[str],
+        mask: _CMap
+    ) -> Iterator[REPPStep]:
+        raise NotImplementedError()
 
 
 class _REPPRule(_REPPOperation):
@@ -190,39 +150,20 @@ class _REPPRule(_REPPOperation):
         self.pattern = pattern
         self.replacement = replacement
         self._re = _compile(pattern)
-
-        groups, literals = parse_template(replacement, self._re)
-        # if a literal is None then it has a group, so make this
-        # easier to iterate over by making pairs of (literal, None) or
-        # (None, group)
-        group_map = dict(groups)
-        self._segments: List[Tuple[Optional[str], Optional[int]]] = [
-            (literal, group_map.get(i))
-            for i, literal
-            in enumerate(literals)
-        ]
-        # either literal or group must be None, but not both
-        assert all((lit is None) != (grp is None)
-                   for lit, grp in self._segments)
-
-        # Get "trackable" capture groups; i.e., those that are
-        # transparent for characterization. For PET behavior, these
-        # must appear in strictly increasing order with no gaps
-        self._last_trackable = -1  # index of trackable segment, not group id
-        last_trackable_group = 0
-        for i, group in groups:
-            if group == last_trackable_group + 1:
-                self._last_trackable = i
-                last_trackable_group = group
-            else:
-                break
+        self._tracked, self._untracked = _get_segments(replacement, self._re)
 
     def __str__(self):
         return f'!{self.pattern}\t\t{self.replacement}'
 
-    def _apply(self, s: str, active: Set[str]) -> Iterator[REPPStep]:
+    def _apply(
+        self,
+        s: str,
+        active: Set[str],
+        mask: _CMap
+    ) -> Iterator[REPPStep]:
         logger.debug(' %s', self)
 
+        applied = False
         ms = list(self._re.finditer(s))
 
         if ms:
@@ -231,95 +172,90 @@ class _REPPRule(_REPPOperation):
             parts: List[str] = []
             smap = array('i', [0])
             emap = array('i', [0])
+            new_mask = array('i', [_MASK_O])
 
             for m in ms:
+                sub, _smap, _emap, _mask, delta, blocked = _process_match(
+                    m, mask, shift, self._tracked, self._untracked
+                )
+                if blocked:
+                    continue
+                applied = True
+
                 start = m.start()
+
+                # copy up to point of match
                 if pos < start:
                     _copy_part(s[pos:start], shift, parts, smap, emap)
+                    new_mask.extend(mask[pos+1:start+1])
 
-                if self._segments:
-                    for literal, start, end, tracked in self._itersegments(m):
-                        if tracked:
-                            _copy_part(literal, shift, parts, smap, emap)
-                        else:
-                            width = end - start
-                            _insert_part(literal, width, shift, parts,
-                                         smap, emap)
-                            shift += width - len(literal)
-                else:
-                    # the replacement is empty (match is deleted)
-                    shift += m.end() - start
-
+                parts.append(sub)
+                smap.extend(_smap)
+                emap.extend(_emap)
+                new_mask.extend(_mask)
+                shift += delta
                 pos = m.end()
 
             if pos < len(s):
                 _copy_part(s[pos:], shift, parts, smap, emap)
+                new_mask.extend(mask[pos+1:len(s)+1])
+
             smap.append(shift)
             emap.append(shift - 1)
+            new_mask.append(_MASK_O)
+            mask = new_mask
             o = ''.join(parts)
-            applied = True
 
         else:
             o = s
             smap = _zeromap(o)
             emap = _zeromap(o)
-            applied = False
 
-        yield REPPStep(s, o, self, applied, smap, emap)
+        yield REPPStep(s, o, self, applied, smap, emap, mask)
 
-    def _itermatches(
-            self, ms: Iterable[Match[str]]
-    ) -> Iterator[Tuple[int, Match[str]]]:
-        """Yield pairs of the last affected position and a match."""
-        last_pos = 0
-        for m in ms:
-            yield (last_pos, m)
-            last_pos = m.end()
 
-    def _itersegments(
-            self, m: Match[str]
-    ) -> Iterator[Tuple[str, int, int, bool]]:
-        """Yield tuples of (replacement, start, end, tracked)."""
-        start = m.start()
+class _REPPMask(_REPPOperation):
+    """
+    A REPP masking rule.
 
-        # first yield segments that might be trackable
-        tracked = self._segments[:self._last_trackable + 1]
-        if tracked:
-            spans = {group: m.span(group)
-                     for _, group in tracked
-                     if group is not None}
-            end = m.start(1)  # if literal before tracked group
-            for literal, group in tracked:
-                if literal is None:
-                    assert group is not None
-                    start, end = spans[group]
-                    yield (m.group(group) or '', start, end, True)
-                    start = end
-                    if group + 1 in spans:
-                        end = spans[group + 1][0]
-                else:
-                    assert literal is not None
-                    yield (literal, start, end, False)
+    When a mask is applied and matches a substring, the substring is
+    blocked from further modification.
 
-        # then group all remaining segments together
-        remaining: List[Optional[str]] = [
-            m.group(grp) if grp is not None else lit
-            for lit, grp in self._segments[self._last_trackable + 1:]
-        ]
-        if remaining:
-            # in some cases m.group(grp) can return None, so replace with ''
-            literal = ''.join(segment or '' for segment in remaining)
-            yield (literal, start, m.end(), False)
+    Args:
+        pattern: the regular expression pattern to match
+    """
+    def __init__(self, pattern: str):
+        self.pattern = pattern
+        self._re = _compile(pattern)
+
+    def __str__(self):
+        return f'={self.pattern}'
+
+    def _apply(
+        self,
+        s: str,
+        active: Set[str],
+        mask: _CMap
+    ) -> Iterator[REPPStep]:
+        logger.debug(' %s', self)
+        newmask = array('i', mask)  # make a copy
+        for m in self._re.finditer(s):
+            start = m.start() + 1
+            newmask[start] = max(_MASK_B, mask[start])
+            for i in range(start + 1, m.end() + 1):
+                newmask[i] = _MASK_I
+        yield REPPStep(s, s, self, True, _zeromap(s), _zeromap(s), newmask)
 
 
 class _REPPGroup(_REPPOperation):
     def __init__(
-            self, operations: List[_REPPOperation] = None, name: str = None
+        self, operations: List[_REPPOperation] = None, name: str = None
     ):
         if operations is None:
             operations = []
         self.operations: List[_REPPOperation] = operations
         self.name = name
+        self._loaded = False
 
     def __repr__(self):
         name = '("{}") '.format(self.name) if self.name is not None else ''
@@ -327,58 +263,52 @@ class _REPPGroup(_REPPOperation):
             type(self).__name__, name, id(self)
         )
 
-    def __str__(self):
-        return 'Module {}'.format(self.name if self.name is not None else '')
-
-    def _apply(self, s: str, active: Set[str]) -> Iterator[REPPStep]:
+    def _apply(
+        self,
+        s: str,
+        active: Set[str],
+        mask: _CMap
+    ) -> Iterator[REPPStep]:
         o = s
         applied = False
         for operation in self.operations:
-            for step in operation._apply(o, active):
+            for step in operation._apply(o, active, mask):
                 yield step
                 o = step.output
+                mask = step.mask
                 applied |= step.applied
 
-        yield REPPStep(s, o, self, applied, _zeromap(o), _zeromap(o))
+        yield REPPStep(s, o, self, applied, _zeromap(o), _zeromap(o), mask)
 
 
-class _REPPGroupCall(_REPPOperation):
-    def __init__(self, name: str, modules: Dict[str, 'REPP']):
-        self.name = name
-        self.modules = modules
-
-    def _apply(self, s: str, active: Set[str]) -> Iterator[REPPStep]:
-        if active is not None and self.name in active:
-            logger.info('>%s', self.name)
-            yield from self.modules[self.name]._apply(s, active)
-            logger.debug('>%s (done)', self.name)
-        else:
-            logger.debug('>%s (inactive)', self.name)
-
-
-class _REPPIterativeGroup(_REPPGroup):
+class _REPPInternalGroup(_REPPGroup):
     def __str__(self):
         return f'Internal group #{self.name}'
 
-    def _apply(self, s: str, active: Set[str]) -> Iterator[REPPStep]:
+    def _apply(
+        self,
+        s: str,
+        active: Set[str],
+        mask: _CMap
+    ) -> Iterator[REPPStep]:
         logger.debug('>%s', self.name)
-        o = s
-        applied = False
-        prev = None
         i = 0
-        while prev != o:
+        prev = s
+        step = None  # in case _REPPGroup._apply() ever yields nothing
+        for step in super()._apply(prev, active, mask):
+            yield step
+            mask = step.mask
+        while step and prev != step.output:
             i += 1
-            prev = o
-            for operation in self.operations:
-                for step in operation._apply(o, active):
-                    yield step
-                    o = step.output
-                    applied |= step.applied
-            yield REPPStep(s, o, self, applied, _zeromap(o), _zeromap(o))
+            prev = step.output
+            for step in super()._apply(prev, active, mask):
+                yield step
+                mask = step.mask
+
         logger.debug('>%s (done; iterated %d time(s))', self.name, i)
 
 
-class REPP(object):
+class REPP(_REPPGroup):
     """
     A Regular Expression Pre-Processor (REPP).
 
@@ -407,22 +337,26 @@ class REPP(object):
     """
 
     def __init__(
-            self,
-            name: str = None,
-            modules: Dict[str, 'REPP'] = None,
-            active: Iterable[str] = None):
+        self,
+        operations: List[_REPPOperation] = None,
+        name: str = None,
+        modules: Dict[str, 'REPP'] = None,
+        active: Iterable[str] = None
+    ):
+        super().__init__(operations=operations, name=name)
         self.info: Optional[str] = None
         self.tokenize_pattern: Optional[str] = None
-        self.group = _REPPGroup(name=name)
 
         if modules is None:
             modules = {}
         self.modules = dict(modules)
+        for modname, mod in self.modules.items():
+            mod.name = modname
         self.active: Set[str] = set()
         if active is None:
             active = []
-        for mod in active:
-            self.activate(mod)
+        for modname in active:
+            self.activate(modname)
 
     @classmethod
     def from_config(cls, path: PathLike, directory=None):
@@ -514,14 +448,9 @@ class REPP(object):
                 activations
         """
         path = Path(path).expanduser()
-        if directory is not None:
-            directory = Path(directory).expanduser()
-        else:
-            directory = path.parent
-        name = path.with_suffix('').name
-        lines = _repp_lines(path)
+        name, directory, lines = _read_file(path, directory)
         r = cls(name=name, modules=modules, active=active)
-        _parse_repp(lines, r, directory)
+        _parse_repp_module(lines, r, directory)
         return r
 
     @classmethod
@@ -537,7 +466,7 @@ class REPP(object):
                 activations
         """
         r = cls(name=name, modules=modules, active=active)
-        _parse_repp(s.splitlines(), r, None)
+        _parse_repp_module(s.splitlines(), r, None)
         return r
 
     def activate(self, mod: str) -> None:
@@ -553,8 +482,20 @@ class REPP(object):
         if mod in self.active:
             self.active.remove(mod)
 
-    def _apply(self, s: str, active: Set[str]) -> Iterator[REPPStep]:
-        return self.group._apply(s, active)
+    def _apply(
+        self,
+        s: str,
+        active: Set[str],
+        mask: _CMap
+    ) -> Iterator[REPPStep]:
+        if self.name in active:
+            logger.info('>%s', self.name)
+            for step in super()._apply(s, active, mask):
+                yield step
+                mask = step.mask
+            logger.debug('>%s (done)', self.name)
+        else:
+            logger.debug('>%s (inactive)', self.name)
 
     def apply(self, s: str, active: Iterable[str] = None) -> REPPResult:
         """
@@ -568,11 +509,10 @@ class REPP(object):
             a :class:`REPPResult` object containing the processed
                 string and characterization maps
         """
-        if active is None:
-            active = self.active
-        else:
-            active = set(active)
-        return self.group.apply(s, active=active)
+        logger.info('apply(%r)', s)
+        active = self.active if active is None else set(active)
+        result = last(self._trace(s, active, False))
+        return result
 
     def trace(
         self, s: str, active: Iterable[str] = None, verbose: bool = False
@@ -591,12 +531,32 @@ class REPP(object):
                 step, and finally a :class:`REPPResult` object after
                 the last rewrite
         """
-        if active is None:
-            active = self.active
-        return self.group.trace(s, active=active, verbose=verbose)
+        logger.info('trace(%r)', s)
+        active = self.active if active is None else set(active)
+        yield from self._trace(s, active, verbose)
+
+    def _trace(
+        self, s: str, active: Set[str], verbose: bool
+    ) -> _Trace:
+        startmap = _zeromap(s)
+        endmap = _zeromap(s)
+        mask = _zeromap(s)
+        # initial boundaries
+        startmap[0] = 1
+        endmap[-1] = -1
+        step = None
+        for step in super()._apply(s, active, mask):
+            if step.applied or verbose:
+                yield step
+            if step.applied:
+                startmap = _mergemap(startmap, step.startmap)
+                endmap = _mergemap(endmap, step.endmap)
+        if step is not None:
+            s = step.output
+        yield REPPResult(s, startmap, endmap)
 
     def tokenize(
-            self, s: str, pattern: str = None, active: Iterable[str] = None
+        self, s: str, pattern: str = None, active: Iterable[str] = None
     ) -> YYTokenLattice:
         """
         Rewrite and tokenize the input string *s*.
@@ -611,17 +571,18 @@ class REPP(object):
             a :class:`~delphin.tokens.YYTokenLattice` containing the
             tokens and their characterization information
         """
+        logger.info('tokenize(%r, %r)', s, pattern)
         if pattern is None:
             if self.tokenize_pattern is None:
                 pattern = DEFAULT_TOKENIZER
             else:
                 pattern = self.tokenize_pattern
-        if active is None:
-            active = self.active
-        return self.group.tokenize(s, pattern=pattern, active=active)
+        active = self.active if active is None else set(active)
+        result = last(self._trace(s, active, False))
+        return self.tokenize_result(result, pattern=pattern)
 
     def tokenize_result(
-            self, result: REPPResult, pattern: str = DEFAULT_TOKENIZER
+        self, result: REPPResult, pattern: str = DEFAULT_TOKENIZER
     ) -> YYTokenLattice:
         """
         Tokenize the result of rule application.
@@ -634,7 +595,14 @@ class REPP(object):
             a :class:`~delphin.tokens.YYTokenLattice` containing the
             tokens and their characterization information
         """
-        return self.group.tokenize_result(result, pattern=pattern)
+        logger.info('tokenize_result(%r, %r)', result, pattern)
+        tokens = [
+            YYToken(id=i, start=i, end=(i + 1),
+                    lnk=Lnk.charspan(tok[0], tok[1]),
+                    form=tok[2])
+            for i, tok in enumerate(_tokenize(result, pattern))
+        ]
+        return YYTokenLattice(tokens)
 
 
 def _compile(pattern: str) -> Pattern[str]:
@@ -653,6 +621,40 @@ def _compile(pattern: str) -> Pattern[str]:
             raise
 
 
+def _get_segments(replacement: str, _re):
+    # parse_template() is an undocumented function in the standard library
+    groups, literals = parse_template(replacement, _re)
+    # literals is a list of strings or None (group position);
+    # groups is a list of (i, g) where i is the index in literals
+    # and g is the capturing group number.
+
+    # first determine the last trackable group, where trackable
+    # segments are transparent for characterization. For PET behavior,
+    # these must appear in strictly increasing order with no gaps
+    last_trackable = 0
+    for expected, (i, grp) in zip(range(1, len(groups)+1), groups):
+        if grp == expected:
+            last_trackable = i + 1  # +1 for slice end
+
+    # we can also combine groups and literals into a single list of
+    # (literal, None) or (None, group) pairs for convenience
+    group_map = dict(groups)
+    segments: List[Tuple[Optional[str], Optional[int]]] = [
+        (literal, group_map.get(i))
+        for i, literal
+        in enumerate(literals)
+    ]
+    # Divide the segments into trackable/untrackable
+    return segments[:last_trackable], segments[last_trackable:]
+
+
+def last(steps: _Trace) -> REPPResult:
+    for step in steps:
+        pass
+    assert isinstance(step, REPPResult)
+    return step
+
+
 def _zeromap(s: str) -> _CMap:
     return array('i', [0] * (len(s) + 2))
 
@@ -671,32 +673,171 @@ def _mergemap(map1: _CMap, map2: _CMap) -> _CMap:
 
 
 def _copy_part(
-        s: str,
-        shift: int,
-        parts: List[str],
-        smap: _CMap,
-        emap: _CMap
+    s: str,
+    shift: int,
+    parts: List[str],
+    smap: _CMap,
+    emap: _CMap
 ) -> None:
     parts.append(s)
-    smap.extend([shift] * len(s))
-    emap.extend([shift] * len(s))
+    map_part = [shift] * len(s)
+    smap.extend(map_part)
+    emap.extend(map_part)
 
 
 def _insert_part(
-        s: str,
-        w: int,
-        shift: int,
-        parts: List[str],
-        smap: _CMap,
-        emap: _CMap
+    s: str,
+    width: int,
+    shift: int,
+    parts: List[str],
+    smap: _CMap,
+    emap: _CMap
 ) -> None:
     parts.append(s)
-    a = shift
-    b = a - len(s)
-    smap.extend(range(a, b, -1))
-    a = shift + w - 1
-    b = a - len(s)
-    emap.extend(range(a, b, -1))
+    endshift = shift - len(s)
+    _width = width - 1
+    smap.extend(range(shift, endshift, -1))
+    emap.extend(range(shift + _width, endshift + _width, -1))
+
+
+def _process_match(
+    m: Match[str],
+    prev_mask: _CMap,
+    shift: int,
+    tracked,
+    untracked
+) -> Tuple[str, _CMap, _CMap, _CMap, int, bool]:
+    parts: List[str] = []
+    smap = array('i', [])
+    emap = array('i', [])
+    mask = array('i', [])
+    delta = 0
+    blocked = False
+
+    if tracked or untracked:
+        start = m.start()
+        # for literals, the end is the start of the next backreference
+        end = next((m.start(g) for _, g in tracked if g), m.end())
+
+        for literal, group in tracked:
+            if literal is None:
+                literal = m.group(group) or ''
+                _copy_part(literal, shift + delta, parts, smap, emap)
+                mask.extend(prev_mask[(start+1):(start+len(literal)+1)])
+                end = m.start(group+1) if group < m.lastindex else m.end()
+            else:
+                # block if overlap with mask
+                if any(prev_mask[start+1:end+1]):
+                    blocked = True
+                    break
+                width = end - start
+                litlen = len(literal)
+                _insert_part(literal, width, shift + delta, parts, smap, emap)
+                mask.extend([_MASK_O] * litlen)
+                delta += width - litlen
+
+            start = end
+
+        if untracked:
+            # block if untracked overlaps with mask, including backreferences
+            if (any(prev_mask[start+1:m.end()+1])
+                or any(any(prev_mask[m.start(grp)+1:m.end(grp)+1])
+                       for lit, grp in untracked if grp)):
+                blocked = True
+            else:
+                # untracked segments can be collapsed into one substring
+                literal = ''.join(
+                    m.group(group) or '' if literal is None else literal
+                    for literal, group in untracked
+                )
+                width = m.end() - start
+                litlen = len(literal)
+                _insert_part(literal, width, shift+delta, parts, smap, emap)
+                mask.extend([_MASK_O] * litlen)
+                delta += width - litlen
+    else:
+        # the replacement is empty (match is deleted)
+        delta = m.end() - m.start()
+
+    substring = ''.join(parts)
+    if not blocked:
+        blocked = _check_mask(substring, m, smap, emap, mask, prev_mask)
+
+    return substring, smap, emap, mask, delta, blocked
+
+
+def _check_mask(
+    s: str,
+    m: Match[str],
+    smap: _CMap,
+    emap: _CMap,
+    mask: _CMap,
+    prev_mask: _CMap
+) -> bool:
+    """Returns True if any masked material has changed.
+
+    There are three contexts for mask checks:
+
+      1. The first character in the match is a mask continuation
+      2. The first character after the end of the match is a mask continuation
+      3. The whole mask is contained within the match
+
+    For (1) and (2), the mask must not change nor move around. For (1)
+    it cannot change but it can move.
+
+    """
+    start, end = m.span()
+    mstart, mend = start + 1, end + 1
+    orig = m.group(0)
+    # no masked material; no problem
+    if not any(prev_mask[mstart:mend]):
+        return False
+    # whether the final mask is fixed depends on the next mask value;
+    # check now as the _make_mask_info() function won't see it
+    end_fixed = prev_mask[mend] == _MASK_I
+    # prev/next left, middle, right
+    pl, pm, pr = _make_mask_info(orig, prev_mask[mstart:mend], end_fixed)
+    nl, nm, nr = _make_mask_info(s, mask, end_fixed)
+
+    # check fixed start/end masks
+    if pl != nl or pr != nr:
+        return True
+    # other masks just need to be present in equal number
+    for substr in set(pm).union(nm):
+        if pm.get(substr, 0) != nm.get(substr, 0):
+            return True
+
+    return False
+
+
+def _make_mask_info(
+    s: str,
+    mask: _CMap,
+    end_fixed: bool
+) -> Tuple[str, Dict[str, int], str]:
+    if not mask:
+        return '', {}, ''
+    left = s[:_get_mask_len(mask, 0)] if mask[0] == _MASK_I else ''
+    right = s[-_get_mask_len(mask[::-1], 0) or len(s):] if end_fixed else ''
+    middle: Dict[str, int] = {}
+    i = len(left)
+    j = len(s) - len(right)
+    while i < j:
+        if mask[i] == _MASK_B:
+            mlen = _get_mask_len(mask, i+1) + 1
+            substr = s[i:mlen]
+            if substr not in middle:
+                middle[substr] = 1
+            else:
+                middle[substr] += 1
+            i += mlen
+        else:
+            i += 1
+    return left, middle, right
+
+
+def _get_mask_len(mask: _CMap, i: int):
+    return sum(1 for _ in takewhile(lambda v: v == _MASK_I, mask[i:]))
 
 
 def _tokenize(result: REPPResult, pattern: str) -> List[Tuple[int, int, str]]:
@@ -716,86 +857,150 @@ def _tokenize(result: REPPResult, pattern: str) -> List[Tuple[int, int, str]]:
     return toks
 
 
+def _read_file(
+    path: Path,
+    directory: Optional[Path]
+) -> Tuple[str, Path, List[str]]:
+    if directory is not None:
+        directory = Path(directory).expanduser()
+    else:
+        directory = path.parent
+    name = path.with_suffix('').name
+    lines = _repp_lines(path)
+    return name, directory, lines
+
+
 def _repp_lines(path: Path) -> List[str]:
     if not path.is_file():
         raise REPPError(f'REPP file not found: {path!s}')
     return path.read_text(encoding='utf-8').splitlines()
 
 
-def _parse_repp(lines: List[str], r: REPP, directory: Path) -> None:
-    ops = list(_parse_repp_group(lines, r, directory))
-    if lines:
-        raise REPPError('Unexpected termination; maybe the # operator '
-                        'appeared without an internal group.')
-    r.group.operations.extend(ops)
+def _parse_repp_module(
+    lines: List[str],
+    r: REPP,
+    directory: Path
+) -> None:
+    r._loaded = True
 
+    operations: List[_REPPOperation] = r.operations
+    stack: List[List[_REPPOperation]] = [operations]
+    internal_groups: Dict[str, _REPPInternalGroup] = {}
 
-def _parse_repp_group(
-    lines: List[str], r: REPP, directory: Path
-) -> Iterator[_REPPOperation]:
-    igs: Dict[str, _REPPIterativeGroup] = {}  # internal groups
     while lines:
         line = lines.pop(0)
         if line.startswith(';') or line.strip() == '':
             continue  # skip comments and empty lines
-        elif line[0] == '!':
-            match = re.match(r'([^\t]+)\t+(.*)', line[1:])
-            if match is None:
-                raise REPPError(f'Invalid rewrite rule: {line}')
-            yield _REPPRule(match.group(1), match.group(2))
-        elif line[0] == '<':
-            fn = directory.joinpath(line[1:].rstrip())
+
+        operator, operand = line[0], line[1:].rstrip()
+
+        if operator == '!':
+            # don't use operand because it was rstripped; use line[1:]
+            operations.append(_parse_rewrite_rule(line[1:]))
+
+        elif operator == '<':
+            fn = directory.joinpath(operand)
             lines = _repp_lines(fn) + lines
-        elif line[0] == '>':
-            modname = line[1:].rstrip()
-            if modname.isdigit():
-                if modname in igs:
-                    yield igs[modname]
-                else:
-                    raise REPPError(
-                        'Iterative group not defined: ' + modname
-                    )
-            else:
-                if modname not in r.modules:
-                    if directory is None:
-                        raise REPPError('Cannot implicitly load modules if '
-                                        'a directory is not given.')
-                    mod = REPP.from_file(
-                        directory.joinpath(modname + '.rpp'),
-                        directory=directory,
-                        modules=r.modules
-                    )
-                    r.modules[modname] = mod
-                yield _REPPGroupCall(modname, r.modules)
-        elif line[0] == '#':
-            igname = line[1:].rstrip()
-            if igname.isdigit():
-                if igname in igs:
-                    raise REPPError(
-                        'Internal group name already defined: ' + igname
-                    )
-                igs[igname] = _REPPIterativeGroup(
-                    operations=list(
-                        _parse_repp_group(lines, r, directory)
-                    ),
-                    name=igname
-                )
-            elif igname == '':
-                return
-            else:
-                raise REPPError('Invalid internal group name: ' + igname)
-        elif line[0] == ':':
-            if r.tokenize_pattern is not None:
-                raise REPPError(
-                    'Only one tokenization pattern (:) may be defined.'
-                )
-            r.tokenize_pattern = line[1:]
-        elif line[0] == '@':
-            if r.info is not None:
-                raise REPPError(
-                    'No more than one meta-info declaration (@) may be '
-                    'defined.'
-                )
-            r.info = line[1:]
+
+        elif operator == '>':
+            operations.append(
+                _handle_group_call(operand, internal_groups, r, directory)
+            )
+
+        elif operator == '=':
+            # don't use operand because it was rstripped; use line[1:]
+            operations.append(_REPPMask(line[1:]))
+
+        elif operator == '#':
+            _handle_internal_group(operand, internal_groups, stack)
+            operations = stack[-1]
+
+        elif operator == ':':
+            _handle_tokenization_pattern(operand, r, len(stack) > 1)
+
+        elif operator == '@':
+            _handle_metainfo_declaration(operand, r, len(stack) > 1)
+
         else:
             raise REPPError(f'Invalid declaration: {line}')
+
+    _verify_internal_groups(internal_groups)
+
+
+def _parse_rewrite_rule(operand: str) -> _REPPRule:
+    match = re.match(r'([^\t]+)\t+(.*)', operand)
+    if match is None:
+        raise REPPError(f'Invalid rewrite rule: !{operand}')
+    return _REPPRule(match.group(1), match.group(2))
+
+
+def _handle_group_call(
+    operand: str,
+    internal_groups,
+    r: REPP,
+    directory: Optional[Path],
+) -> _REPPGroup:
+    if operand.isdigit():
+        if operand not in internal_groups:
+            internal_groups[operand] = _REPPInternalGroup(operations=[],
+                                                          name=operand)
+        return internal_groups[operand]
+    elif not operand:
+        raise REPPError('Missing group name')
+    else:
+        if operand not in r.modules:
+            r.modules[operand] = REPP(name=operand, modules=r.modules)
+        mod = r.modules[operand]
+        if not mod._loaded:
+            if directory is None:
+                raise REPPError('Cannot implicitly load modules if '
+                                'a directory is not given.')
+            modpath = directory / (operand + '.rpp')
+            _parse_repp_module(_repp_lines(modpath), mod, directory)
+        return mod
+
+
+def _handle_internal_group(operand, internal_groups, stack) -> None:
+    if operand.isdigit():
+        if operand not in internal_groups:
+            internal_groups[operand] = _REPPInternalGroup(operations=[],
+                                                          name=operand)
+        ig = internal_groups[operand]
+        if ig._loaded:
+            raise REPPError(f'Internal group name already defined: {operand}')
+        ig._loaded = True
+        stack.append(ig.operations)
+    elif operand == '':
+        stack.pop()
+    else:
+        raise REPPError('Invalid internal group name: ' + operand)
+
+
+def _handle_tokenization_pattern(
+    operand: str,
+    r: REPP,
+    in_internal_group: bool,
+) -> None:
+    if in_internal_group:
+        raise REPPError('tokenization pattern defined in internal group')
+    if r.tokenize_pattern is not None:
+        raise REPPError('Only one tokenization pattern (:) may be defined.')
+    r.tokenize_pattern = operand
+
+
+def _handle_metainfo_declaration(
+    operand: str,
+    r: REPP,
+    in_internal_group: bool,
+) -> None:
+    if in_internal_group:
+        raise REPPError('meta-info declaration defined in internal group')
+    if r.info is not None:
+        raise REPPError('Only one meta-info declaration (@) may be defined.')
+    r.info = operand
+
+
+def _verify_internal_groups(internal_groups):
+    for grpname, grp in internal_groups.items():
+        if not grp._loaded:
+            raise REPPError(f'internal group not defined: #{grpname}')
